/* ---------------------------------------------------------------------------------------------------------------------------- */
/* CREATE DB USERS as specified by Table
	FOR SPECIFIED DATABASE*/
/* CREATE THE DATABASE BEFORE EXECUTING THIS SCRIPT */
/* This procedure adds or replaces DB Users, granting each role membership in OpI Security Roles. 
	so that DB Users made Members will inherit those permissions */

/* SELECT INTENDED DATABASE, IN SSMS, MSVisualStudio, etc., VIA DROPDOWN IN MENU BAR OR BY SPECIFYING DATABASE WHEN CONNECTING FILE TO RUN  */

/* If an error is encountered at execution of this SQL file, stating that "setproperty must be called on the UI thread, 
	that can be fixed by repairing OLEDB using its installer. There may be other configuration methods to repair this, 
	but they are currently not identified. */

/* If error is reported in Message stating unable to initialize ACE OLEDB, make sure to close the Excel source Table doc. */

/* ---------------------------------------------------------------------------------------------------------------------------- */
/* Security Architecture:  
	https://learn.microsoft.com/en-us/sql/relational-databases/security/authentication-access/getting-started-with-database-engine-permissions?view=sql-server-ver16
	https://www.mssqltips.com/sqlservertip/2699/sql-server-user-defined-server-roles/ */

/* ---------------------------------------------------------------------------------------------------------------------------- */

/* Entire work Copyright (C) 2022 Richard M. Bixler, All Rights Reserved. Licensing Terms Stated Below. */

/* ---------------------------------------------------------------------------------------------------------------------------- */

/* Extract specifications and sequence to GRANT DB User-Role Permissions */
/* From Table rngOpIUsers specifying DB Users and OpI Role in the selected Database */
SET NOCOUNT ON; 

DECLARE @usrLineNo AS INT
DECLARE @usrUserName AS NVARCHAR(max) = NULL
DECLARE @usrScope as NVARCHAR(max) = NULL
DECLARE @usrIdentifierText AS NVARCHAR(max) = NULL
DECLARE	@usrIdentifierToken AS NVARCHAR(max) = NULL
DECLARE @usrActiveNow AS NVARCHAR(max) = NULL
DECLARE	@usrOpIVerb AS NVARCHAR(max) = NULL
DECLARE	@usrRoleName AS NVARCHAR(max) = NULL
DECLARE	@usrRoleVerb AS NVARCHAR(max) = NULL
DECLARE @usrOnEnvironment AS NVARCHAR(max) = NULL
DECLARE	@usrDatabase AS NVARCHAR(max) = NULL

/* From Table rngOpIRolesGRANT specifying Permissions to grant to role */
DECLARE @grLineNo AS INT
DECLARE @grRole AS NVARCHAR(max) = NULL
DECLARE @grScope as NVARCHAR(max) = NULL
DECLARE @grVerb AS NVARCHAR(max) = NULL
DECLARE @grGrantOpt AS NVARCHAR(max) = NULL
DECLARE	@grPermission AS NVARCHAR(max) = NULL
DECLARE @grOnObject AS NVARCHAR(max) = NULL
DECLARE	@grObjectName AS NVARCHAR(max) = NULL
DECLARE	@grPerModifier2 AS NVARCHAR(max) = NULL
DECLARE @grAS AS NVARCHAR(max) = NULL

/* Local Variables */
DECLARE @currentDatabase AS NVARCHAR(max) = NULL
DECLARE @currentUserName as NVARCHAR(max) = NULL
DECLARE	@query AS NVARCHAR(max) = NULL
DECLARE	@DBWrapQuery AS NVARCHAR(max) = NULL
DECLARE @OnObj AS NVARCHAR(max) = NULL
DECLARE @member_principalID AS NVARCHAR(max) = NULL
DECLARE @memberCount AS INT = 0
DECLARE @usrRoleNameFixed AS NVARCHAR(max) = NULL
DECLARE @memberName AS NVARCHAR(max) = NULL
DECLARE @roleName AS NVARCHAR(max) = NULL
DECLARE @createOrAlter AS NVARCHAR(max) = NULL

/* Local for permission grant */
DECLARE @roleDesc as NVARCHAR(max) = NULL
DECLARE @roleType as NVARCHAR(max) = NULL
DECLARE @phraseAS as NVARCHAR(max) = NULL

/* CURSOR Blocks */
DECLARE @OpIMemberCount AS INT = 0

/* TRY...CATCH Blocks */
DECLARE @Plural as NVARCHAR(1) = N''
DECLARE @errorNumber as INT = 0
DECLARE @errorMessage as NVARCHAR(max) = NULL
DECLARE @errorLine as INT = 0

PRINT N'SECURITY_5-CREATE-DB-Users.sql'; PRINT N'';

SET @currentDatabase = DB_NAME();	-- Where are we?

DECLARE usercreate_query_cursor CURSOR FOR 
	SELECT usrLineNo, usrUserName, usrScope, usrIdentifierText, usrIdentifierToken, usrActiveNow, usrOpIVerb, usrRoleVerb, usrRoleName, usrOnEnvironment, usrDatabase
		FROM LS_OpISQLServerSecurity_ToolSource...rngOpIUsers AS xlOpIUsersCreate
		
		WHERE (xlOpIUsersCreate.usrScope = 'DB' 
			AND NOT(xlOpIUsersCreate.usrActiveNow = 'No')
			AND xlOpIUsersCreate.usrOpIVerb IS NOT NULL
			AND xlOpIUsersCreate.usrDatabase = @currentDatabase)

		ORDER BY 
			xlOpIUsersCreate.usrLineNo,
			xlOpIUsersCreate.usrUserName;

OPEN usercreate_query_cursor;
FETCH NEXT FROM usercreate_query_cursor 
	INTO @usrLineNo, @usrUserName, @usrScope, @usrIdentifierText, @usrIdentifierToken, @usrActiveNow, @usrOpIVerb, @usrRoleVerb, @usrRoleName, @usrOnEnvironment, @usrDatabase;

WHILE @@FETCH_STATUS = 0 
BEGIN
	IF @usrActiveNow IN (N'Yes', N'DryRun')
	BEGIN
		PRINT N''; 
		IF @usrActiveNow = N'DryRun' PRINT N'Dry Run';
		PRINT CONCAT(N'Row ', @usrLineNo, N' ', @usrOpIVerb, N' ', @usrUserName, N' ', @usrDatabase);
		
		IF ( @usrOpIVerb = N'DROP' )
		BEGIN
			/* Drop previously-set OpI DB User-Role User-Role memberships set up separately. */
			IF EXISTS (SELECT [sys].[database_principals].[name], [sys].[database_principals].[type] 
				FROM [sys].[database_principals] 
				WHERE [type] IN (N'U', N'S') AND [name] = @usrUserName)
			BEGIN
				/* Drop known memberships by this @usrUserName in this @usrDatabase */
				/* Get all roleNames for @usrUserName, into temp table #OpIUserMembershipTable */
				DROP TABLE IF EXISTS #OpIUserMembershipTable;	-- Just in case, to avoid any mistake
				SELECT roleMemberNames.name AS memberName, rolePrincipals.name AS roleName
					INTO #OpIUserMembershipTable
					FROM [sys].[database_principals] AS roleMemberNames
					LEFT OUTER JOIN [sys].[database_role_members] AS roleMembers
						ON roleMemberNames.principal_id = roleMembers.member_principal_id
					LEFT OUTER JOIN [sys].[database_principals] AS rolePrincipals
						ON roleMembers.role_principal_id = rolePrincipals.principal_id
					WHERE
						(roleMemberNames.type IN (N'S')
						AND rolePrincipals.principal_id < 1000
						AND roleMemberNames.name = @usrUserName
						AND rolePrincipals.type IN (N'R'));

				/* See if we caught any */
				SELECT @OpIMemberCount = COUNT(*) FROM (SELECT * FROM #OpIUserMembershipTable) AS x;
				IF NOT (@OpIMemberCount = 1) SET @Plural = N's' ELSE SET @Plural = N'';
				PRINT CONCAT(N'User "', @usrUserName, N'" has ', @OpIMemberCount, N' role membership', @Plural);

				/* Use a Cursor to step through #OpIUserMembershipTable to retrieve @roleName : @memberName pairs to DROP */
				DECLARE OpIUserMembership_cursor CURSOR FOR (SELECT #OpIUserMembershipTable.roleName, #OpIUserMembershipTable.memberName FROM #OpIUserMembershipTable);
				OPEN OpIUserMembership_cursor;
				FETCH NEXT FROM OpIUserMembership_cursor 
					INTO @roleName, @memberName;

				WHILE @@FETCH_STATUS = 0
				BEGIN
					/* Drop Role memberships */
					IF EXISTS (SELECT [sys].[database_principals].[name], [sys].[database_principals].[type] 
						FROM [sys].[database_principals] 
						WHERE [type] IN (N'R') AND [name] = @roleName)
					BEGIN
						SET @errorNumber = 0;
						BEGIN TRY
							SET @query = CONCAT(N'ALTER ROLE ', ' ', @roleName, N' DROP MEMBER ', @memberName, N';');
							IF @usrActiveNow = N'Yes' EXEC sp_executesql @query;
						END TRY
						BEGIN CATCH
							SELECT @errorNumber = ERROR_NUMBER(), @errorMessage = ERROR_MESSAGE(), @errorLine = ERROR_LINE();
						END CATCH
						PRINT @query;
						IF NOT @errorNumber = 0 PRINT CONCAT(N'ERROR ', @errorNumber, ', Line ', @errorLine, ' ', @errorMessage);
					END;
					
					FETCH NEXT FROM OpIUserMembership_cursor 
						INTO @roleName, @memberName;
				END;

				/* Clean up Role-Membership drop loop by dropping temporary table and associated cursor */
				CLOSE OpIUserMembership_cursor;
				DEALLOCATE OpIUserMembership_cursor;
				DROP TABLE IF EXISTS #OpIUserMembershipTable;

				/* DROP this OpI DB User. */
				BEGIN
					BEGIN TRY
						SET @errorNumber = 0;
						SET @query = CONCAT(N'DROP USER IF EXISTS ', @usrUserName, N';');	
						IF @usrActiveNow = N'Yes' EXEC sp_executesql @query;
					END TRY
					BEGIN CATCH
						SELECT @errorNumber = ERROR_NUMBER(), @errorMessage = ERROR_MESSAGE(), @errorLine = ERROR_LINE();
					END CATCH
					PRINT @query;
					IF NOT @errorNumber = 0 PRINT CONCAT(N'ERROR ', @errorNumber, ', Line ', @errorLine, ' ', @errorMessage);
				END;
			END;
		END;

		IF ( @usrOpIVerb = N'ADD/UPDATE' )
		BEGIN
			/* Create a clean instance of this OpI DB User if it does not exist */
			IF NOT EXISTS (SELECT [sys].[database_principals].[name], [sys].[database_principals].[type] 
				FROM [sys].[database_principals] 
				WHERE [type] IN (N'U', N'S') AND [name] = @usrUserName)
			BEGIN 
				IF EXISTS (SELECT [sys].[sql_logins].[name], [sys].[sql_logins].[type] 
					FROM [sys].[sql_logins] 
					WHERE [type] IN (N'U', N'S') AND [name] = @usrUserName) 
				BEGIN
					/* OpI DB User Name follows LoginName */
					BEGIN TRY
						SET @errorNumber = 0;
						SET @query = CONCAT(N'CREATE USER [', @usrUserName, '] FROM LOGIN [', @usrUserName, N'];'); 
						IF @usrActiveNow = N'Yes' EXEC sp_executesql @query;
					END TRY
					BEGIN CATCH
						SELECT @errorNumber = ERROR_NUMBER(), @errorMessage = ERROR_MESSAGE(), @errorLine = ERROR_LINE();
					END CATCH
					PRINT @query;
					IF NOT @errorNumber = 0 PRINT CONCAT(N'ERROR ', @errorNumber, ', Line ', @errorLine, ' ', @errorMessage);
				END;
				ELSE PRINT CONCAT(N'User "', @usrUserName, '" not created, Login "', @usrUserName, N'" does not exist');
			END;
			ELSE PRINT CONCAT(N'User "', @usrUserName, N'" already exists');
		END;

		IF ( @usrOpIVerb = N'ADD/UPDATE' )
		BEGIN
			/* Add DB User GRANT * and * FIXED OpI Database-Role membership */
			/* First check that the DB-User Name has been created from an existing Server-Login Name */
			IF EXISTS (SELECT [sys].[database_principals].[name], [sys].[database_principals].[type] 
				FROM [sys].[database_principals] 
				WHERE [type] IN (N'U', N'S') AND [name] = @usrUserName) 
			BEGIN
				SET @usrRoleNameFixed = CONCAT(@usrRoleName, N'Fixed');
				IF (@usrRoleVerb = N'ADD' OR @usrRoleVerb = N'DROP')
				BEGIN
					/* ADD/DROP membership to GRANT-Role Permissions Role for the database if that role exists */
					IF EXISTS (SELECT [sys].[database_principals].[name], [sys].[database_principals].[type] 
						FROM [sys].[database_principals] 
						WHERE [type] IN (N'R') AND [name] = @usrRoleName)
					BEGIN
						BEGIN TRY
							SET @errorNumber = 0;
							SET @query = CONCAT(N'ALTER ROLE ', @usrRoleName, ' ', @usrRoleVerb, N' MEMBER ', @usrUserName, N';');
							IF @usrActiveNow = N'Yes' EXEC sp_executesql @query;
						END TRY
						BEGIN CATCH
							SELECT @errorNumber = ERROR_NUMBER(), @errorMessage = ERROR_MESSAGE(), @errorLine = ERROR_LINE();
						END CATCH
						PRINT @query;
						IF NOT @errorNumber = 0 PRINT CONCAT(N'ERROR ', @errorNumber, ', Line ', @errorLine, ' ', @errorMessage);
					END;
				END;
			END;

			/* Add permissions from tblOpIRolesGRANT classed as "DBUser" */
			IF EXISTS (SELECT [sys].[database_principals].[name], [sys].[database_principals].[type] 
				FROM [sys].[database_principals] 
				WHERE [type] IN (N'U', N'S') AND [name] = @usrUserName) 
			BEGIN 				
			/* Step Through tblOpIRolesGRANT and add roles classed as "DBUser" */
			/* Inner-Loop to GRANT permissions to Role created or to be modified */
				DECLARE grant_cursor CURSOR FOR 
					SELECT grLineNo, grRole, grScope, grVerb, usrDatabase, grPermission, grOnObject, grObjectName, grGrantOpt, grAS
			
					FROM LS_OpISQLServerSecurity_ToolSource...rngOpIUsers AS xlOpIUsersCreate
						LEFT OUTER JOIN LS_OpISQLServerSecurity_ToolSource...rngOpIRolesGRANT AS xlOpIRolesGRANT 
							ON xlOpIUsersCreate.usrUserName = xlOpIRolesGRANT.grRole

						WHERE (xlOpIUsersCreate.usrScope = 'DB' 
							AND xlOpIRolesGRANT.grScope = 'DB'
							AND xlOpIRolesGRANT.grRoleClass = 'DBUser'
							AND xlOpIRolesGRANT.grVerb IS NOT NULL
							AND NOT(xlOpIUsersCreate.usrActiveNow = 'No')
							AND xlOpIRolesGRANT.grRole = @usrRoleName
							AND xlOpIUsersCreate.usrDatabase = @currentDatabase)

						ORDER BY 
							xlOpIUsersCreate.usrLineNo, 
							xlOpIRolesGRANT.grLineNo;

				OPEN grant_cursor;
				FETCH NEXT FROM grant_cursor 
					INTO @grLineNo, @grRole, @grScope, @grVerb, @usrDatabase, @grPermission, @grOnObject, @grObjectName, @grGrantOpt, @grAS;

				WHILE @@FETCH_STATUS = 0  
				BEGIN
					/* Configure database-level permissions to the role using GRANT, DENY, REVOKE: https://learn.microsoft.com/en-us/sql/t-sql/statements/create-role-transact-sql?view=sql-server-ver16 */
					/* To invoke this, in tblOpIRolesGRANT add rows specifying permissions to the user defined by tblOpIUsers.usrUserName
						specifying that @usrUserName in tblOpIRoleGrant.grRole, and  column .grRoleClass as "DBUser". Add rows specifying permissions,
						and those marked GRANT, REVOKE or DENY will be executed for that user. */
						
					SET @onObj = NULL;
					SELECT @onObj = [xlOpIOnObj].[onObjectName]
						FROM [LS_OpISQLServerSecurity_ToolSource]...[rngOpIOnObj] AS xlOpIOnObj
						WHERE ( [xlOpIOnObj].[onEnvironment] = @usrOnEnvironment 
							AND [xlOpIOnObj].[onOnObjName] = @grObjectName );
					IF @onObj = N'onDatabase' SET @onObj = @usrDatabase;

					SET @grPerModifier2 = NULL;
					IF @grGrantOpt = N'Yes'
					BEGIN
						IF @grVerb = N'GRANT'	SET @grPerModifier2 = N' WITH GRANT OPTION';
						IF @grVerb = N'REVOKE'	SET @grPerModifier2 = N' CASCADE';
						IF @grVerb = N'DENY'	SET @grPerModifier2 = N' CASCADE';
					END;

					IF @grOnObject = N'ON' SET @grOnObject = N'ON ';

					IF @grAS IS NOT NULL
					BEGIN
						SET @phraseAS = NULL;
						SELECT @phraseAS = [xlOpIOnObj].[onObjectName]
							FROM [LS_OpISQLServerSecurity_ToolSource]...[rngOpIOnObj] AS xlOpIOnObj
							WHERE ( [xlOpIOnObj].[onEnvironment] = @usrOnEnvironment 
								AND [xlOpIOnObj].[onOnObjName] = @grAs );
						IF @phraseAS = N'onDatabase' SET @phraseAS = @usrDatabase ELSE SET @phraseAS = @grAS;
						SET @phraseAs = CONCAT(N' AS ', @phraseAS);
					END;
						
					SET @errorNumber = 0;
					BEGIN TRY
						SET @query = CONCAT(@grVerb, ' ', @grPermission, ' ', @grOnObject, @onObj, ' TO [', @usrUserName, ']', @grPerModifier2, @phraseAS, ';');
						IF @usrActiveNow = N'Yes' EXEC sp_executesql @query;
					END TRY
					BEGIN CATCH
						SELECT @errorNumber = ERROR_NUMBER(), @errorMessage = ERROR_MESSAGE(), @errorLine = ERROR_LINE();
					END CATCH
					PRINT @query;
					IF NOT @errorNumber = 0 PRINT CONCAT(N'ERROR ', @errorNumber, ', rolLineNo ', @usrLineNo, ', GrLineNo ', @grLineNo, ', ', @errorMessage);

					FETCH NEXT FROM grant_cursor 
						INTO @grLineNo, @grRole, @grScope, @grVerb, @usrDatabase, @grPermission, @grOnObject, @grObjectName, @grGrantOpt, @grAS;
				END;

				CLOSE grant_cursor;
				DEALLOCATE grant_cursor;
			END;
		END;

		IF ( @usrOpIVerb = N'ALTER' )
		BEGIN
			/* Insert ALTER USER code here */
			PRINT N'';
			PRINT N'ALTER';
			PRINT CONCAT(N'User "', @usrUserName, N'" already exists');
		END;
	END;

	FETCH NEXT FROM usercreate_query_cursor 
		INTO @usrLineNo, @usrUserName, @usrScope, @usrIdentifierText, @usrIdentifierToken, @usrActiveNow, @usrOpIVerb, @usrRoleVerb, @usrRoleName, @usrOnEnvironment, @usrDatabase;

END;

CLOSE usercreate_query_cursor;
DEALLOCATE usercreate_query_cursor;

/* LICENSING
Power Operational Intelligence ("PowerOpI") Framework
   Copyright (C) 2020 Richard M. Bixler
Derived products MSAccessTool, PowerQueryTool, SQLServerTool, and SQLServerRemote
   Copyright � 2021 Richard M. Bixler
Derived product OpISQLServerSecurity
   Copyright � 2022 Richard M. Bixler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or 
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

Contact:
rmbpublic@comcast.net
www.softtoyssoftware.com */