/* ---------------------------------------------------------------------------------------------------------------------------- */
/* CREATE SVR ROLES OpIRoleSvrSuper, OpIRoleSvrManager, OpIRoleSvrDataOwner, OpIRoleSvrClient
	FOR SERVER INSTANCE <SERVERNAME>\<INSTANCENAME> FOR ASSIGNMENT OF PERMISSIONS TO EACH SERVER LOGIN */
/* ---------------------------------------------------------------------------------------------------------------------------- */

/* Security Architecture:  
	https://learn.microsoft.com/en-us/sql/relational-databases/security/authentication-access/getting-started-with-database-engine-permissions?view=sql-server-ver16
	https://www.mssqltips.com/sqlservertip/2699/sql-server-user-defined-server-roles/ 
	https://learn.microsoft.com/en-us/sql/relational-databases/system-catalog-views/sys-server-role-members-transact-sql?view=sql-server-ver16 */

/* If an error is encountered at execution of this SQL file, stating that "setproperty must be executed in the UI thread, 
	that can be fixed by repairing OLEDB using its installer. There may be other configuration methods to repair this, 
	but they are currently not identified. */

/* If error is reported in Message stating unable to initialize ACE OLEDB, make sure to close the Excel source Table doc. */

/* ---------------------------------------------------------------------------------------------------------------------------- */
	
/* Entire work Copyright (C) 2022 Richard M. Bixler, All Rights Reserved. Licensing Terms Stated Below. */

/* ---------------------------------------------------------------------------------------	*/
/* Extract specifications and sequence to GRANT Server-Login Role Permissions */

/* SERVER ROLES are created and managed in built-in database "Master", */
/*	so externally-set DB choice is overridden. */
USE master;
GO
SET NOCOUNT ON; 

/* From Table xlOpIRolesCreate specifying management of Roles */
DECLARE @rolLineNo AS INT
DECLARE @rolScope AS VARCHAR(10)
DECLARE @rolFixedGrant AS NVARCHAR(15)
DECLARE	@rolActiveNow AS NVARCHAR(6) = N'No'
DECLARE @rolOpIVerb AS NVARCHAR(max) = NULL
DECLARE	@rolDatabase AS NVARCHAR(max) = NULL
DECLARE	@rolRoleName AS NVARCHAR(max) = NULL
DECLARE @rolMemberOfVerb AS NVARCHAR(max) = NULL
DECLARE @rolMemberOfRole AS NVARCHAR(max) = NULL
DECLARE @rolOnEnvironment AS NVARCHAR(max) = NULL

/* From Table rngOpIRolesGRANT specifying Permissions to grant to role */
DECLARE @grLineNo AS INT
DECLARE @grRole AS NVARCHAR(max) = NULL
DECLARE @grScope as NVARCHAR(max) = NULL
DECLARE @grVerb AS NVARCHAR(max) = NULL
DECLARE @grGrantOpt AS NVARCHAR(max) = NULL
DECLARE	@grPermission AS NVARCHAR(max) = NULL
DECLARE @grOnObject AS NVARCHAR(max) = NULL
DECLARE	@grObjectName AS NVARCHAR(max) = NULL
DECLARE	@grPerModifier2 AS NVARCHAR(max) = NULL
DECLARE @grAS AS NVARCHAR(max) = NULL

/* Local Variables */
DECLARE @currentDatabase AS NVARCHAR(max) = NULL
DECLARE	@query AS NVARCHAR(max) = NULL
DECLARE	@DBWrapQuery AS NVARCHAR(max) = NULL
DECLARE @OnObj AS NVARCHAR(max) = NULL
DECLARE @member_principalID AS NVARCHAR(max) = NULL
DECLARE @memberCount as INT = 0
DECLARE @roleDesc as NVARCHAR(max) = NULL
DECLARE @roleType as NVARCHAR(max) = NULL
DECLARE @phraseAS as NVARCHAR(max) = NULL

/* CURSOR Blocks */
DECLARE @OpIMemberCount AS INT = 0
DECLARE @roleName as NVARCHAR(max) = NULL
Declare @memberName as NVARCHAR(max) = NULL

/* TRY...CATCH Blocks */
DECLARE @Plural as NVARCHAR(1) = N''
DECLARE @errorNumber as INT = 0
DECLARE @errorMessage as NVARCHAR(max) = NULL
DECLARE @errorLine as INT = 0

SET @currentDatabase = DB_NAME();	-- Where are we? Must be Master for Svr Role permissions.
IF @currentDatabase = N'master'
BEGIN
	PRINT N'SECURITY_1-CREATE-Svr-GRANT-Roles.sql'; PRINT N'';
	/* Grant EXECUTE to OLEDB_Provider */
	-- GRANT EXECUTE ON SYS.XP_PROP_OLEDB_PROVIDER TO [Public];
	-- Databases > System Databases > master > Programmability > Extended Stored Procedures > System Extended Stored Procedures > sys.xp_prop_oledb_provider
	-- This GRANT is handled in tblOpIRolesGRANT, invoked from  tblOpIRolesCreate

	/* Step through the OpIRolesCreate table and grant permissions as defined in the OpIRolesGRANT table  */
	DECLARE rolecreate_query_cursor CURSOR FOR 
		SELECT rolLineNo, rolScope, rolFixedGrant, rolActiveNow, rolOpIVerb, rolDatabase, rolRoleName, rolMemberOfVerb, rolMemberOfRole, rolOnEnvironment
		FROM LS_OpISQLServerSecurity_ToolSource...rngOpIRolesCreate AS xlOpIRolesCreate
		
			WHERE (xlOpIRolesCreate.rolScope = 'Svr' 
				AND xlOpIRolesCreate.rolFixedGrant = 'Grant' 
				AND NOT(xlOpIRolesCreate.rolActiveNow = 'No')
				AND xlOpIRolesCreate.rolDatabase = @currentDatabase)

			ORDER BY 
				xlOpIRolesCreate.rolLineNo;

	OPEN rolecreate_query_cursor;
	FETCH NEXT FROM rolecreate_query_cursor 
		INTO @rolLineNo, @rolScope, @rolFixedGrant, @rolActiveNow, @rolOpIVerb, @rolDatabase, @rolRoleName, @rolMemberOfVerb, @rolMemberOfRole, @rolOnEnvironment;

	WHILE @@FETCH_STATUS = 0 
	BEGIN
		IF @rolActiveNow IN (N'Yes', N'DryRun')
		BEGIN
			PRINT N'';
			IF @rolActiveNow = N'DryRun' PRINT N'Dry Run';
			PRINT CONCAT(N'Row ', @rolLineNo, N' ', @rolScope, N' ', @rolFixedGrant, N' ', @rolOpIVerb, N' ', @rolDatabase, N' ', @rolRoleName);

			IF ( @rolOpIVerb = N'DROP' )
			BEGIN
				IF EXISTS (SELECT [sys].[server_principals].[name], [sys].[server_principals].[type] 
					FROM [sys].[server_principals] 
					WHERE [type] IN (N'R') AND [name] = @rolRoleName)
				BEGIN
					/* Drop previously-set memberships, then members, of this @rolRoleName in this @rolDatabase */
					/* ------------------------------------------------------------------------------ */
					/* FIRST, DROP MEMBERSHIPS of this OpI Server User-Role in other User-Role memberships 
						set up separately. */
					/* ------------------------------------------------------------------------------ */
					/* Get all roleNames for @RoleName, into temp table #OpIRoleMembershipTable */
					DROP TABLE IF EXISTS #OpIRoleMembershipTable;	-- Just in case, to avoid any mistake
					SELECT roleMemberNames.name AS memberName, rolePrincipals.name AS roleName
						INTO #OpIRoleMembershipTable
						FROM [sys].[server_principals] AS roleMemberNames
						LEFT OUTER JOIN [sys].[server_role_members] AS roleMembers
							ON roleMemberNames.principal_id = roleMembers.member_principal_id -- member_principal_id to drop MEMBERSHIPS, role_principal_id to drop MEMBERS
						LEFT OUTER JOIN [sys].[server_principals] AS rolePrincipals
							ON roleMembers.role_principal_id = rolePrincipals.principal_id
						WHERE
							(roleMemberNames.name = @rolRoleName
							AND roleMemberNames.type IN (N'R')
							AND NOT(rolememberNames.name = rolePrincipals.name)
							-- AND rolePrincipals.principal_id < 1000	--Not needed?
							AND rolePrincipals.type IN (N'R'));

					/* See if we caught any */
					SELECT @OpIMemberCount = COUNT(*) FROM (SELECT * FROM #OpIRoleMembershipTable) AS x;
					IF NOT (@OpIMemberCount = 1) SET @Plural = N's' ELSE SET @Plural = N'';
					PRINT CONCAT(N'User ', @rolRoleName, N' has ', @OpIMemberCount, N' role membership', @Plural);

					/* Use a Cursor to step through #OpIRoleMembershipTable to retrieve @roleName : @memberName pairs to DROP */
					DECLARE OpIRoleMembership_cursor CURSOR FOR (SELECT #OpIRoleMembershipTable.roleName, #OpIRoleMembershipTable.memberName FROM #OpIRoleMembershipTable);
					OPEN OpIRoleMembership_cursor;
					FETCH NEXT FROM OpIRoleMembership_cursor 
						INTO @roleName, @memberName;
						/* @roleName contains the name of the Role whose memberships we're dropping */
						/* @memberName contains the name of a Role in which Role @roleName is a member */
						/* We will drop Role @roleName from membership in Role @memberName */

						-- THIS LOGIC IS TO DROP MEMBERSHIPS
						-- MAY NEED TO SWAP USE OF @roleNAME and @memberName, for logic to to drop MEMBERS.

					WHILE @@FETCH_STATUS = 0
					BEGIN
						/* Drop Role memberships */
						IF EXISTS (SELECT [sys].[server_principals].[name], [sys].[server_principals].[type]
							FROM [sys].[server_principals] 
							WHERE [type] IN (N'R') AND [name] = @memberName)
						BEGIN
							SELECT @roleDesc=[sys].[server_principals].[type_desc]
								FROM [sys].[server_principals] 
								WHERE [sys].[server_principals].[type] IN (N'R') AND [sys].[server_principals].[name] = @memberName;
							IF @roleDesc = N'SERVER_ROLE' SET @roleType = N'SERVER ' ELSE SET @roleType = N'';
						
							SET @errorNumber = 0;
							BEGIN TRY
								SET @query = CONCAT(N'ALTER ', @roleType, 'ROLE ', @roleName, N' DROP MEMBER ', @memberName, N';');
								IF @rolActiveNow = N'Yes' EXEC sp_executesql @query;
							END TRY
							BEGIN CATCH
								SELECT @errorNumber = ERROR_NUMBER(), @errorMessage = ERROR_MESSAGE(), @errorLine = ERROR_LINE();
							END CATCH
							PRINT @query;
							IF NOT @errorNumber = 0 PRINT CONCAT(N'ERROR ', @errorNumber, ', Line ', @errorLine, ' ', @errorMessage);
						END;
					
						FETCH NEXT FROM OpIRoleMembership_cursor 
							INTO @roleName, @memberName;
					END;

					/* Clean up Role-Membership drop loop by dropping temporary table and associated cursor */
					CLOSE OpIRoleMembership_cursor;
					DEALLOCATE OpIRoleMembership_cursor;
					DROP TABLE IF EXISTS #OpIRoleMembershipTable;

					/* ------------------------------------------------------------------------------ */
					/* THEN, DROP MEMBERS of this OpI Server User-Role @rolRoleName. */
					/* ------------------------------------------------------------------------------ */
					/* Get all memberNames for @RoleName, into temp table #OpIRoleMemberTable */
					DROP TABLE IF EXISTS #OpIRoleMemberTable;	-- Just in case, to avoid any mistake
					SELECT roleMemberNames.name AS memberName, rolePrincipals.name AS roleName
						INTO #OpIRoleMemberTable
						FROM [sys].[server_principals] AS roleMemberNames
						LEFT OUTER JOIN [sys].[server_role_members] AS roleMembers
							ON roleMemberNames.principal_id = roleMembers.role_principal_id -- member_principal_id to drop MEMBERSHIPS, role_principal_id to drop MEMBERS
						LEFT OUTER JOIN [sys].[server_principals] AS rolePrincipals
							ON roleMembers.member_principal_id = rolePrincipals.principal_id
						WHERE
							(roleMemberNames.name = @rolRoleName
							AND roleMemberNames.type IN (N'R')
							AND NOT(rolememberNames.name = rolePrincipals.name)
							AND rolePrincipals.principal_id < 1000	--Not needed?
							AND rolePrincipals.type IN (N'S', N'U', N'R'));

					/* See if we caught any */
					SELECT @OpIMemberCount = COUNT(*) FROM (SELECT * FROM #OpIRoleMemberTable) AS x;
					IF NOT (@OpIMemberCount = 1) SET @Plural = N's' ELSE SET @Plural = N'';
					PRINT CONCAT(N'User ', @rolRoleName, N' has ', @OpIMemberCount, N' role member', @Plural);

					/* Use a Cursor to step through #OpIRoleMembershipTable to retrieve @roleName : @memberName pairs to DROP */
					DECLARE OpIRoleMember_cursor CURSOR FOR (SELECT #OpIRoleMemberTable.roleName, #OpIRoleMemberTable.memberName FROM #OpIRoleMemberTable);
					OPEN OpIRoleMember_cursor;
					FETCH NEXT FROM OpIRoleMember_cursor 
						INTO @roleName, @memberName;
						/* @roleName contains the name of the Role whose memberships we're dropping */
						/* @memberName contains the name of a Role in which Role @roleName is a member */
						/* We will drop Role @roleName from membership in Role @memberName */

						-- THIS LOGIC IS TO DROP MEMBERS
						-- MAY NEED TO SWAP USE OF @roleNAME and @memberName, for logic to drop MEMBERS.

					WHILE @@FETCH_STATUS = 0
					BEGIN
						/* Drop Role members */
						IF EXISTS (SELECT [sys].[server_principals].[name], [sys].[server_principals].[type]
							FROM [sys].[server_principals] 
							WHERE [type] IN (N'S', N'U', N'R') AND [name] = @memberName)
						BEGIN
							SELECT @roleDesc=[sys].[server_principals].[type_desc]
								FROM [sys].[server_principals] 
								WHERE [sys].[server_principals].[type] IN (N'S', N'U', N'R') AND [sys].[server_principals].[name] = @memberName;
							IF @roleDesc = N'SERVER_ROLE' SET @roleType = N'SERVER ' ELSE SET @roleType = N'';
						
							SET @errorNumber = 0;
							BEGIN TRY
								SET @query = CONCAT(N'ALTER ', @roleType, 'ROLE ', @memberName, N' DROP MEMBER ', @roleName, N';');
								IF @rolActiveNow = N'Yes' EXEC sp_executesql @query;
							END TRY
							BEGIN CATCH
								SELECT @errorNumber = ERROR_NUMBER(), @errorMessage = ERROR_MESSAGE(), @errorLine = ERROR_LINE();
							END CATCH
							PRINT @query;
							IF NOT @errorNumber = 0 PRINT CONCAT(N'ERROR ', @errorNumber, ', Line ', @errorLine, ' ', @errorMessage);
						END;
					
						FETCH NEXT FROM OpIRoleMember_cursor 
							INTO @roleName, @memberName;
					END;

					/* Clean up Role-Member drop loop by dropping temporary table and associated cursor */
					CLOSE OpIRoleMember_cursor;
					DEALLOCATE OpIRoleMember_cursor;
					DROP TABLE IF EXISTS #OpIRoleMemberTable;

					/* ------------------------------------------------------------------------------ */
					/* FINALLY, DROP this OpI Server Role @rolRoleName when memberships and 
						members of Role @rolRoleName have been DROPPED. */
					/* ------------------------------------------------------------------------------ */
					BEGIN
						BEGIN TRY
							IF @rolScope = N'SVR' SET @roleDesc = N'SERVER' ELSE SET @roleDesc = N''; -- Drop Role @rolRoleName after memberships and members have been dropped
							SET @errorNumber = 0;
							SET @query = CONCAT(N'DROP ', @roleDesc, ' ROLE [', @rolRoleName, N'];');
							IF @rolActiveNow = N'Yes' EXEC sp_executesql @query;
						END TRY
						BEGIN CATCH
							SELECT @errorNumber = ERROR_NUMBER(), @errorMessage = ERROR_MESSAGE(), @errorLine = ERROR_LINE();
						END CATCH
						PRINT @query;
						IF NOT @errorNumber = 0 PRINT CONCAT(N'ERROR ', @errorNumber, ', Line ', @errorLine, ' ', @errorMessage);
					END;
				END;
			END;

			IF ( @rolOpIVerb = N'ADD/UPDATE' )
			BEGIN
				/* Create, or re-create, a clean instance of this OpI DB User Role */
				IF NOT EXISTS (SELECT [sys].[server_principals].[name], [sys].[server_principals].[type] 
					FROM [sys].[server_principals] 
					WHERE [type] IN (N'R') AND [name] = @rolRoleName)
				BEGIN 
					SET @errorNumber = 0;
					BEGIN TRY
						SET @query = CONCAT(N'CREATE SERVER ROLE [', @rolRoleName, '] AUTHORIZATION sysadmin;'); 
						IF @rolActiveNow = N'Yes' EXEC sp_executesql @query;
					END TRY
					BEGIN CATCH
						SELECT @errorNumber = ERROR_NUMBER(), @errorMessage = ERROR_MESSAGE(), @errorLine = ERROR_LINE();
					END CATCH
					PRINT @query;
					IF NOT @errorNumber = 0 PRINT CONCAT(N'ERROR ', @errorNumber, ', Line ', @errorLine, ' ', @errorMessage);
				END;
				ELSE PRINT CONCAT(N'SERVER ROLE ', @rolRoleName, N' already exists.');
			END;

			IF ( @rolOpIVerb = N'ADD/UPDATE' )
			/* Inner-Loop to GRANT permissions to Role created or to be modified */
			BEGIN
				DECLARE query_cursor CURSOR FOR 
					SELECT grLineNo, grRole, grScope, grVerb, rolDatabase, grPermission, grOnObject, grObjectName, grGrantOpt, grAS
			
					FROM LS_OpISQLServerSecurity_ToolSource...rngOpIRolesCreate AS xlOpIRolesCreate
						LEFT OUTER JOIN LS_OpISQLServerSecurity_ToolSource...rngOpIRolesGRANT AS xlOpIRolesGRANT 
							ON xlOpIRolesCreate.rolRoleName = xlOpIRolesGRANT.grRole

						WHERE (xlOpIRolesCreate.rolScope = 'Svr' 
							AND xlOpIRolesGRANT.grScope = 'Svr'
							AND NOT(xlOpIRolesGRANT.grRoleClass IN('SvrLogin', 'DBUser'))
							AND xlOpIRolesCreate.rolFixedGrant = 'Grant' 
							AND xlOpIRolesGRANT.grVerb IS NOT NULL
							AND NOT(xlOpIRolesCreate.rolActiveNow = 'No')
							AND xlOpIRolesGRANT.grRole = @rolRoleName
							AND xlOpIRolesCreate.rolDatabase = @currentDatabase)

						ORDER BY 
							xlOpIRolesCreate.rolLineNo, 
							xlOpIRolesGRANT.grLineNo;

				OPEN query_cursor;
				FETCH NEXT FROM query_cursor 
					INTO @grLineNo, @grRole, @grScope, @grVerb, @rolDatabase, @grPermission, @grOnObject, @grObjectName, @grGrantOpt, @grAS;

				WHILE @@FETCH_STATUS = 0  
				BEGIN
					IF EXISTS (SELECT [sys].[server_principals].[name], [sys].[server_principals].[type] 
						FROM [sys].[server_principals] 
						WHERE [type] IN (N'R') AND [name] = @grRole)
					BEGIN
						/* Configure database-level permissions to the role using GRANT, DENY, REVOKE: https://learn.microsoft.com/en-us/sql/t-sql/statements/create-role-transact-sql?view=sql-server-ver16 */
						/* Database Fixed Role db_owner */
						SET @onObj = NULL;
						SELECT @onObj = [xlOpIOnObj].[onObjectName]
							FROM [LS_OpISQLServerSecurity_ToolSource]...[rngOpIOnObj] AS xlOpIOnObj
							WHERE ( [xlOpIOnObj].[onEnvironment] = @rolOnEnvironment 
								AND [xlOpIOnObj].[onOnObjName] = @grObjectName );
						IF @onObj = N'onDatabase' SET @onObj = @rolDatabase;

						SET @grPerModifier2 = NULL;
						IF @grGrantOpt = N'Yes'
						BEGIN
							IF @grVerb = N'GRANT'	SET @grPerModifier2 = N' WITH GRANT OPTION';
							IF @grVerb = N'REVOKE'	SET @grPerModifier2 = N' CASCADE';
							IF @grVerb = N'DENY'	SET @grPerModifier2 = N' CASCADE';
						END;

						IF @grOnObject = N'ON' SET @grOnObject = N'ON ';

						IF @grAS IS NOT NULL
						BEGIN
							SET @phraseAS = NULL;
							SELECT @phraseAS = [xlOpIOnObj].[onObjectName]
								FROM [LS_OpISQLServerSecurity_ToolSource]...[rngOpIOnObj] AS xlOpIOnObj
								WHERE ( [xlOpIOnObj].[onEnvironment] = @rolOnEnvironment 
									AND [xlOpIOnObj].[onOnObjName] = @grAs );
							IF @phraseAS = N'onDatabase' SET @phraseAS = N'sysadmin' ELSE SET @phraseAS = @grAS;
							SET @phraseAs = CONCAT(N' AS ', @phraseAS);
						END;
						
						SET @errorNumber = 0;
						BEGIN TRY
							SET @query = CONCAT(@grVerb, ' ', @grPermission, ' ', @grOnObject, @onObj, ' TO [', @grRole, ']', @grPerModifier2, @phraseAS, ';');
							IF @rolActiveNow = N'Yes' EXEC sp_executesql @query;
						END TRY
						BEGIN CATCH
							SELECT @errorNumber = ERROR_NUMBER(), @errorMessage = ERROR_MESSAGE(), @errorLine = ERROR_LINE();
						END CATCH
						PRINT @query;
						IF NOT @errorNumber = 0 PRINT CONCAT(N'ERROR ', @errorNumber, ', rolLineNo ', @rolLineNo, ', GrLineNo ', @grLineNo, ', ', @errorMessage);
					END;

					FETCH NEXT FROM query_cursor 
						INTO @grLineNo, @grRole, @grScope, @grVerb, @rolDatabase, @grPermission, @grOnObject, @grObjectName, @grGrantOpt, @grAS;
				END;

				CLOSE query_cursor;
				DEALLOCATE query_cursor;

				/* ADD/DROP membership by current-line @rolRoleName into specified Upstream Role @rolMemberOfRole if that role exists */
				/* Upstream role @rolMemberOfRole is not constrained to be an OpI role. */
				IF EXISTS (SELECT [sys].[database_principals].[name], [sys].[database_principals].[type] 
				FROM [sys].[database_principals] 
				WHERE [type] IN (N'R') AND [name] = @rolMemberOfRole) 
				BEGIN
					IF (@rolMemberOfVerb = N'ADD' OR @rolMemberOfVerb = N'DROP')	-- Ignore upstream ADD/DROP unless @rolMemberOfVerb specifies that action
					BEGIN
						BEGIN TRY
							SET @errorNumber = 0;
							SET @query = CONCAT(N'ALTER ROLE ', @rolMemberOfRole, ' ', @rolMemberOfVerb, N' MEMBER ', @rolRoleName, N';');
							IF @rolActiveNow = N'Yes' EXEC sp_executesql @query;
						END TRY
						BEGIN CATCH
							SELECT @errorNumber = ERROR_NUMBER(), @errorMessage = ERROR_MESSAGE(), @errorLine = ERROR_LINE();
						END CATCH
						PRINT @query;
						IF NOT @errorNumber = 0 PRINT CONCAT(N'ERROR ', @errorNumber, ', Line ', @errorLine, ' ', @errorMessage);
					END;
				END;
			END;

			FETCH NEXT FROM rolecreate_query_cursor 
				INTO @rolLineNo, @rolScope, @rolFixedGrant, @rolActiveNow, @rolOpIVerb, @rolDatabase, @rolRoleName, @rolMemberOfVerb, @rolMemberOfRole, @rolOnEnvironment;
		END;
	END;

	CLOSE rolecreate_query_cursor;
	DEALLOCATE rolecreate_query_cursor;
END;

/* LICENSING
Power Operational Intelligence ("PowerOpI") Framework
   Copyright (C) 2020 Richard M. Bixler
Derived products MSAccessTool, PowerQueryTool, SQLServerTool, and SQLServerRemote
   Copyright  2021 Richard M. Bixler
Derived product OpISQLServerSecurity
   Copyright  2022 Richard M. Bixler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or 
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

Contact:
rmbpublic@comcast.net
www.softtoyssoftware.com */