/* ---------------------------------------------------------------------------------------------------------------------------- */
/* CREATE Server Logins for roles OpIRoleSvrSuper, OpIRoleSvrManager, OpIRoleSvrDataOwner, OpIRoleSvrClient */
/* Logins exist at the SERVER level, so adds/updates/deletes are made to database "master". That override
	is provided in this SQL source, so it is not necessary to select "master" in the database connection toolbar. */

/* This procedure adds or updates Server Logins, granting role membership in OpI Svr Security Roles to each Login 
	so that server logins made Members will inherit those permissions */

/* If an error is encountered at execution of this SQL file, stating that "setproperty must be executed in the UI thread, 
	that can be fixed by repairing OLEDB using its installer. There may be other configuration methods to repair this, 
	but they are currently not identified. */

/* If error is reported in Message stating unable to initialize ACE OLEDB, make sure to close the Excel source Table doc. */

/* ---------------------------------------------------------------------------------------------------------------------------- */
/* Security Architecture:  
	https://learn.microsoft.com/en-us/sql/relational-databases/security/authentication-access/getting-started-with-database-engine-permissions?view=sql-server-ver16
	https://www.mssqltips.com/sqlservertip/2699/sql-server-user-defined-server-roles/ 
	https://learn.microsoft.com/en-us/sql/t-sql/statements/create-login-transact-sql?view=sql-server-ver16 */

/* ---------------------------------------------------------------------------------------------------------------------------- */

/* Entire work Copyright (C) 2022 Richard M. Bixler, All Rights Reserved. Licensing Terms Stated Below. */

USE master;
GO
SET NOCOUNT ON; 

/* Extract specifications and sequence to GRANT DB User-Role Permissions */
/* From Table rngOpIUsers specifying DB Users and OpI Role in the selected Database */
DECLARE @linLineNo AS INT
DECLARE @linActiveNow AS NVARCHAR(max) = NULL
DECLARE	@linOpIVerb AS NVARCHAR(max) = NULL
DECLARE	@linOpILoginWindows AS NVARCHAR(max) = NULL
DECLARE	@linLoginDomain AS NVARCHAR(max) = NULL
DECLARE @linLoginName AS NVARCHAR(max) = NULL
DECLARE @linIdentifierText AS NVARCHAR(max) = NULL
DECLARE	@linIdentifierToken AS NVARCHAR(max) = NULL
DECLARE	@linSetPassword AS NVARCHAR(max) = NULL
DECLARE	@linHashed AS NVARCHAR(max) = NULL
DECLARE	@linMustChange AS NVARCHAR(max) = NULL
DECLARE	@linPasswordExpires AS NVARCHAR(max) = NULL
DECLARE	@linCheckPolicy AS NVARCHAR(max) = NULL
DECLARE	@linRoleVerb AS NVARCHAR(max) = NULL
DECLARE	@linRoleName AS NVARCHAR(max) = NULL
DECLARE	@linSID AS NVARCHAR(max) = NULL
DECLARE	@linSvrCredentialVerb AS NVARCHAR(max) = NULL
DECLARE	@linSvrCredential AS NVARCHAR(max) = NULL
DECLARE	@linWinCertificate AS NVARCHAR(max) = NULL
DECLARE	@linAsymmetricKey AS NVARCHAR(max) = NULL
DECLARE	@linDefaultDatabase AS NVARCHAR(max) = NULL
DECLARE	@linDefaultLanguage AS NVARCHAR(max) = NULL
DECLARE	@linLoginStatus AS NVARCHAR(max) = NULL
DECLARE	@linOldPassword AS NVARCHAR(max) = NULL
DECLARE	@linNewLoginName AS NVARCHAR(max) = NULL
DECLARE	@linEKMCredentialVerb AS NVARCHAR(max) = NULL
DECLARE	@linEKMCredentialName AS NVARCHAR(max) = NULL
DECLARE	@linOnEnvironment AS NVARCHAR(max) = NULL
DECLARE	@linDatabase AS NVARCHAR(max) = NULL
DECLARE @linScope AS NVARCHAR(max) = NULL
DECLARE @linFullLoginName AS NVARCHAR(max) = NULL

/* From Table rngOpIRolesGRANT specifying Permissions to grant to role */
DECLARE @grLineNo AS INT
DECLARE @grRole AS NVARCHAR(max) = NULL
DECLARE @grScope as NVARCHAR(max) = NULL
DECLARE @grVerb AS NVARCHAR(max) = NULL
DECLARE @grGrantOpt AS NVARCHAR(max) = NULL
DECLARE	@grPermission AS NVARCHAR(max) = NULL
DECLARE @grOnObject AS NVARCHAR(max) = NULL
DECLARE	@grObjectName AS NVARCHAR(max) = NULL
DECLARE	@grPerModifier2 AS NVARCHAR(max) = NULL
DECLARE @grAS AS NVARCHAR(max) = NULL

/* Local Variables */
DECLARE @currentDatabase AS NVARCHAR(max) = NULL
DECLARE @currentUserName as NVARCHAR(max) = NULL
DECLARE	@with NVARCHAR(max) = NULL
DECLARE	@query AS NVARCHAR(max) = NULL
DECLARE @linRoleNameFixed AS NVARCHAR(max) = NULL
DECLARE @passwordOptions AS NVARCHAR(max) = NULL
DECLARE @SQLLoginOptions AS NVARCHAR(max) = NULL
DECLARE @status_options AS NVARCHAR(max) = NULL
DECLARE @set_options AS NVARCHAR(max) = NULL
DECLARE @cryptographic_credentials_option AS NVARCHAR(max) = NULL
DECLARE @password_option AS NVARCHAR(max) = NULL
DECLARE @defaultLanguageDB AS NVARCHAR(max) = NULL

/* Local for permission grant */
DECLARE	@DBWrapQuery AS NVARCHAR(max) = NULL
DECLARE @OnObj AS NVARCHAR(max) = NULL
DECLARE @member_principalID AS NVARCHAR(max) = NULL
DECLARE @memberCount as INT = 0
DECLARE @roleDesc as NVARCHAR(max) = NULL
DECLARE @roleType as NVARCHAR(max) = NULL
DECLARE @phraseAS as NVARCHAR(max) = NULL

/* CURSOR Blocks */
DECLARE @OpIMemberCount AS INT = 0
DECLARE @memberName AS NVARCHAR(max) = NULL
DECLARE @roleName AS NVARCHAR(max) = NULL

/* TRY...CATCH Blocks */
DECLARE @Plural as NVARCHAR(1) = N''
DECLARE @errorNumber as INT = 0
DECLARE @errorMessage as NVARCHAR(max) = NULL
DECLARE @errorLine as INT = 0

PRINT N'SECURITY_2-CREATE-Svr-Logins.sql'; PRINT N'';

SET @currentDatabase = DB_NAME();	-- Where are we?

DECLARE logincreate_query_cursor CURSOR FOR 
	SELECT linLineNo, linActiveNow, linOpIVerb, linOpILoginWindows, linLoginDomain, linLoginName, linIdentifierText, linIdentifierToken, linSetPassword, linHashed, linMustChange, linPasswordExpires, linCheckPolicy, linRoleVerb, linRoleName, linSID, linSvrCredentialVerb, linSvrCredential, linWinCertificate, linAsymmetricKey, linDefaultDatabase, linDefaultLanguage, linLoginStatus, linOldPassword, linNewLoginName, linEKMCredentialVerb, linEKMCredentialName, linOnEnvironment, linDatabase, linScope, linFullLoginName
		FROM LS_OpISQLServerSecurity_ToolSource...rngOpILogins AS xlOpILoginsCreate
		
		WHERE (xlOpILoginsCreate.linScope = 'Svr' 
			AND NOT (xlOpILoginsCreate.linActiveNow = 'No')
			AND xlOpILoginsCreate.linOpIVerb IS NOT NULL)

		ORDER BY 
			xlOpILoginsCreate.linLineNo,
			xlOpILoginsCreate.linLoginName;

OPEN logincreate_query_cursor;
FETCH NEXT FROM logincreate_query_cursor 
	INTO @linLineNo, @linActiveNow, @linOpIVerb, @linOpILoginWindows, @linLoginDomain, @linLoginName, @linIdentifierText, @linIdentifierToken, @linSetPassword, @linHashed, @linMustChange, @linPasswordExpires, @linCheckPolicy, @linRoleVerb, @linRoleName, @linSID, @linSvrCredentialVerb, @linSvrCredential, @linWinCertificate, @linAsymmetricKey, @linDefaultDatabase, @linDefaultLanguage, @linLoginStatus, @linOldPassword, @linNewLoginName, @linEKMCredentialVerb, @linEKMCredentialName, @linOnEnvironment, @linDatabase, @linScope, @linFullLoginName;

WHILE @@FETCH_STATUS = 0 
BEGIN
	IF @linActiveNow IN (N'Yes', N'DryRun')
	BEGIN
		PRINT N'';
		IF @linActiveNow = N'DryRun' PRINT N'Dry Run';
		PRINT CONCAT(N'Row ', @linLineNo, N' ', @linOpIVerb, N' ', @linOpILoginWindows, N' ', @linLoginName);

		IF ( @linOpIVerb = N'DROP' )
		BEGIN
			/* Drop previously-set OpI-Login Server-Role memberships set up separately. */
			IF EXISTS (SELECT [sys].[sql_logins].[name], [sys].[sql_logins].[type] 
				FROM [sys].[sql_logins] 
				WHERE [type] IN (N'U', N'S') AND [name] = @linLoginName)
			BEGIN
				/* Drop known memberships by this @linLoginName in this @usrDatabase */
				/* Get all roleNames for @linLoginName, into temp table #OpILoginMembershipTable */
				DROP TABLE IF EXISTS #OpILoginMembershipTable;	-- Just in case, to avoid any mistake
				SELECT roleMemberNames.name AS memberName, /*roleMemberNames.type AS roleMemberType, roleMembers.member_principal_id,*/ rolePrincipals.name AS roleName/*, rolePrincipals.type, rolePrincipals.principal_id*/
					INTO #OpILoginMembershipTable
					FROM [sys].[server_principals] AS roleMemberNames					-- Could use [sys].[sql_logins] just as well
					LEFT OUTER JOIN [sys].[server_role_members] AS roleMembers
						ON roleMemberNames.principal_id = roleMembers.member_principal_id
					LEFT OUTER JOIN [sys].[server_principals] AS rolePrincipals			-- Could use [sys].[sql_logins] just as well
						ON roleMembers.role_principal_id = rolePrincipals.principal_id

					WHERE
						(roleMemberNames.type IN (N'S')
						AND rolePrincipals.principal_id < 1000
						AND roleMemberNames.name = @linLoginName
						AND rolePrincipals.type IN (N'R'));

				/* See if we caught any */
				SELECT @OpIMemberCount = COUNT(*) FROM (SELECT * FROM #OpILoginMembershipTable) AS x;
				IF NOT (@OpIMemberCount = 1) SET @Plural = N's' ELSE SET @Plural = N'';
				PRINT CONCAT(N'Login "', @linLoginName, N'" has ', @OpIMemberCount, N' role membership', @Plural);

				/* Use a Cursor to step through ##OpILoginMembershipTable to retrieve @roleName : @memberName pairs to DROP */
				DECLARE OpILoginMembership_cursor CURSOR FOR (SELECT #OpILoginMembershipTable.roleName, #OpILoginMembershipTable.memberName FROM #OpILoginMembershipTable);
				OPEN OpILoginMembership_cursor;
				FETCH NEXT FROM OpILoginMembership_cursor 
					INTO @roleName, @memberName;

				WHILE @@FETCH_STATUS = 0
				BEGIN
					/* Drop Role memberships by this OpI Login. Table #OpILoginMembershipTable was populated
					with constraint that field @memberName in the Table are equal to @linLoginName; and a record
					exists in the Table for each @roleName in which this @memberName (=linLoginName) is a member. 
					Those Role memberships in OpILoginMembershipTable will now be DROPPED. */
					IF EXISTS (SELECT [sys].[server_principals].[name], [sys].[server_principals].[type] 
						FROM [sys].[server_principals] 
						WHERE [type] IN (N'R') AND [name] = @roleName)
					BEGIN
						SET @errorNumber = 0;
						BEGIN TRY
							SET @query = CONCAT(N'ALTER SERVER ROLE ', ' ', @roleName, N' DROP MEMBER ', @memberName, N';');
							IF @linActiveNow = N'Yes' EXEC sp_executesql @query;
						END TRY
						BEGIN CATCH
							SELECT @errorNumber = ERROR_NUMBER(), @errorMessage = ERROR_MESSAGE(), @errorLine = ERROR_LINE();
						END CATCH
						PRINT @query;
						IF NOT @errorNumber = 0 PRINT CONCAT(N'ERROR ', @errorNumber, ', Line ', @errorLine, ' ', @errorMessage);
					END;
					
					FETCH NEXT FROM OpIUserMembership_cursor 
						INTO @roleName, @memberName;
				END;

				/* Clean up Role-Membership drop loop by dropping temporary table and associated cursor */
				CLOSE OpILoginMembership_cursor;
				DEALLOCATE OpILoginMembership_cursor;
				DROP TABLE IF EXISTS #OpILoginMembershipTable;

				/* When memberships of this OpI Login into other Roles are cleared, DROP this OpI Login. */
				BEGIN
					SET @errorNumber = 0;
					BEGIN TRY
						SET @query = CONCAT(N'DROP LOGIN ', @linLoginName, N';');	
						IF @linActiveNow = N'Yes' EXEC sp_executesql @query;
					END TRY
					BEGIN CATCH
							SELECT @errorNumber = ERROR_NUMBER(), @errorMessage = ERROR_MESSAGE(), @errorLine = ERROR_LINE();
					END CATCH
					PRINT @query;
					IF NOT @errorNumber = 0 PRINT CONCAT(N'ERROR ', @errorNumber, ', Line ', @errorLine, ' ', @errorMessage);
				END;
			END;
		END;

		IF ( @linOpIVerb = N'ADD/UPDATE' )
		BEGIN
			IF NOT EXISTS (SELECT [sys].[sql_logins].[name], [sys].[sql_logins].[type] 
				FROM [sys].[sql_logins] 
				WHERE [type] IN (N'U', N'S') AND [name] = @linLoginName) 
			BEGIN 				
				/* Create a clean instance of this OpI Server Login if it does not exist */
				/* WINDOWS LOGIN */
				IF @linOpILoginWindows = N'Windows'
				BEGIN
					SET @with = NULL;
					IF @linDefaultDatabase IS NOT NULL SET @with = CONCAT(N'DEFAULT_DATABASE = ', @linDefaultDatabase);
					
					IF @with IS NOT NULL AND @linDefaultLanguage IS NOT NULL SET @with = @with + ', ';
					IF @linDefaultLanguage IS NOT NULL SET @with = CONCAT(@with, N' DEFAULT_LANGUAGE = ', @linDefaultLanguage);

					IF @with IS NOT NULL SET @With = CONCAT(N' WITH ', @with)

					IF @linLoginDomain IS NOT NULL SET @linLoginName = @linLoginDomain + N'\' + @linLoginName;

					SET @query = CONCAT(N'CREATE LOGIN [', @linFullLoginName, N'] FROM WINDOWS', @with, N';');
				END;
				
				/* CERTIFICATE LOGIN */
				IF @linOpILoginWindows = N'Certificate'
				BEGIN
					SET @with = N' FROM CERTIFICATE ';
					SET @query = CONCAT(N'CREATE LOGIN [', @linLoginName, N'] ', @with, @linWinCertificate, N';');
				END;

				/* ASYMMETRIC KEY LOGIN */
				IF @linOpILoginWindows = N'Asymmetric Key'
				BEGIN
					SET @with = N' FROM ASYMMETRIC KEY ';
					SET @query = CONCAT(N'CREATE LOGIN [', @linLoginName, N'] ', @with, @linAsymmetricKey, N';');
				END;

				/* SQL SERVER LOGIN */
				IF @linOpILoginWindows IN(N'SQLServer', N'Credential')
				BEGIN
					SET @with = NULL;

					SET @passwordOptions = NULL;

					IF @linSetPassword IS NOT NULL SET @passwordOptions = CONCAT(N'PASSWORD = ', N'''', @linSetPassword, N'''');
					IF @linHashed = N'Yes' SET @passwordOptions = @passwordOptions + N' HASHED';

					IF @passwordOptions IS NOT NULL AND @linMustChange = N'MUST_CHANGE' SET @passwordOptions = @passwordOptions + ', ';
					IF @linMustChange = N'MUST_CHANGE' SET @passwordOptions = @passwordOptions + N' MUST_CHANGE';

					IF @passwordOptions IS NOT NULL AND @linPasswordExpires IN(N'ON', N'OFF') SET @passwordOptions = @passwordOptions + ', ';
					IF @linPasswordExpires IN(N'ON', N'OFF') SET @passwordOptions = @passwordOptions + N' CHECK_EXPIRATION = ' + @linPasswordExpires;
					
					IF @passwordOptions IS NOT NULL AND @linCheckPolicy IN(N'ON', N'OFF') SET @passwordOptions = @passwordOptions + ', ';
					IF @linCheckPolicy IN(N'ON', N'OFF') SET @passwordOptions = @passwordOptions + N' CHECK_POLICY = ' + @linCheckPolicy;


					SET @SQLLoginOptions = NULL;
					IF @linSID IS NOT NULL SET @SQLLoginOptions = N'SID = ' + @linSID;

					IF @SQLLoginOptions IS NOT NULL AND @linDefaultDatabase IS NOT NULL SET @SQLLoginOptions = @SQLLoginOptions + ', ';
					IF @linDefaultDatabase IS NOT NULL SET @SQLLoginOptions = @SQLLoginOptions + N' DEFAULT_DATABASE = ' + @linDefaultDatabase;

					IF @SQLLoginOptions IS NOT NULL AND @linDefaultLanguage IS NOT NULL SET @SQLLoginOptions = @SQLLoginOptions + ', ';
					IF @linDefaultLanguage IS NOT NULL SET @SQLLoginOptions = @SQLLoginOptions + N' DEFAULT_LANGUAGE = ' + @linDefaultLanguage;

					IF @SQLLoginOptions IS NOT NULL AND @linSvrCredential IS NOT NULL SET @SQLLoginOptions = @SQLLoginOptions + ', ';
					IF @linSvrCredential IS NOT NULL SET @SQLLoginOptions = @SQLLoginOptions + N' CREDENTIAL = ' + @linSvrCredential;

					IF @passwordOptions IS NOT NULL AND @SQLLoginOptions IS NOT NULL SET @passwordOptions = @passwordOptions + N', ';
					SET @with = CONCAT(N' WITH ', @passwordOptions, @SQLLoginOptions)
					SET @query = CONCAT(N'CREATE LOGIN [', @linLoginName, N'] ', @with, N';');
				END;

				SET @errorNumber = 0;
				PRINT N'';
				PRINT @query;
				BEGIN TRY
					IF @linActiveNow = N'Yes' EXEC sp_executesql @query;
				END TRY
				BEGIN CATCH
					SELECT @errorNumber = ERROR_NUMBER(), @errorMessage = ERROR_MESSAGE(), @errorLine = ERROR_LINE();
				END CATCH
				/* Query was already printed to log creation. */
				IF NOT @errorNumber = 0 PRINT CONCAT(N'ERROR ', @errorNumber, ', Line ', @errorLine, ' ', @errorMessage);
			END;
			
			/* Add permissions from tblOpIRolesGRANT classed as "SvrLogin" */
			IF EXISTS (SELECT [sys].[sql_logins].[name], [sys].[sql_logins].[type] 
				FROM [sys].[sql_logins] 
				WHERE [type] IN (N'U', N'S') AND [name] = @linLoginName) 
			BEGIN 				
			/* Step Through tblOpIRolesGRANT and add roles classed as "SvrLogin" */
			/* Inner-Loop to GRANT permissions to Role created or to be modified */
				DECLARE grant_cursor CURSOR FOR 
					SELECT grLineNo, grRole, grScope, grVerb, linDatabase, grPermission, grOnObject, grObjectName, grGrantOpt, grAS
			
					FROM LS_OpISQLServerSecurity_ToolSource...rngOpILogins AS xlOpILoginsCreate
						LEFT OUTER JOIN LS_OpISQLServerSecurity_ToolSource...rngOpIRolesGRANT AS xlOpIRolesGRANT 
							ON xlOpILoginsCreate.linRoleName = xlOpIRolesGRANT.grRole

						WHERE (xlOpILoginsCreate.linScope = 'Svr' 
							AND xlOpIRolesGRANT.grScope = 'Svr'
							AND xlOpIRolesGRANT.grRoleClass = 'SvrLogin'	-- grRoleClass = "SvrLogin" will attempt GRANT to linLoginName rather than to grRole.
							AND xlOpIRolesGRANT.grVerb IS NOT NULL
							AND NOT(xlOpILoginsCreate.linActiveNow = 'No')
							AND xlOpIRolesGRANT.grRole = @linRoleName
							AND xlOpILoginsCreate.linDatabase = @currentDatabase)

						ORDER BY 
							xlOpILoginsCreate.linLineNo, 
							xlOpIRolesGRANT.grLineNo;

				OPEN grant_cursor;
				FETCH NEXT FROM grant_cursor 
					INTO @grLineNo, @grRole, @grScope, @grVerb, @linDatabase, @grPermission, @grOnObject, @grObjectName, @grGrantOpt, @grAS;

				WHILE @@FETCH_STATUS = 0  
				BEGIN
					/* Configure database-level permissions to the role using GRANT, DENY, REVOKE: https://learn.microsoft.com/en-us/sql/t-sql/statements/create-role-transact-sql?view=sql-server-ver16 */
					/* Database Fixed Role db_owner */
						
					IF (@linOpILoginWindows = N'Windows' AND @linLoginDomain IS NOT NULL) SET @linLoginName = CONCAT(@linLoginDomain, N'\', @linLoginName);
					
					SET @onObj = NULL;
					SELECT @onObj = [xlOpIOnObj].[onObjectName]
						FROM [LS_OpISQLServerSecurity_ToolSource]...[rngOpIOnObj] AS xlOpIOnObj
						WHERE ( [xlOpIOnObj].[onEnvironment] = @linOnEnvironment 
							AND [xlOpIOnObj].[onOnObjName] = @grObjectName );
					IF @onObj = N'onDatabase' SET @onObj = @linDatabase;

					SET @grPerModifier2 = NULL;
					IF @grGrantOpt = N'Yes'
					BEGIN
						IF @grVerb = N'GRANT'	SET @grPerModifier2 = N' WITH GRANT OPTION';
						IF @grVerb = N'REVOKE'	SET @grPerModifier2 = N' CASCADE';
						IF @grVerb = N'DENY'	SET @grPerModifier2 = N' CASCADE';
					END;

					IF @grOnObject = N'ON' SET @grOnObject = N'ON ';

					IF @grAS IS NOT NULL
					BEGIN
						SET @phraseAS = NULL;
						SELECT @phraseAS = [xlOpIOnObj].[onObjectName]
							FROM [LS_OpISQLServerSecurity_ToolSource]...[rngOpIOnObj] AS xlOpIOnObj
							WHERE ( [xlOpIOnObj].[onEnvironment] = @linOnEnvironment 
								AND [xlOpIOnObj].[onOnObjName] = @grAs );
						IF @phraseAS = N'onDatabase' SET @phraseAS = @linDatabase ELSE SET @phraseAS = @grAS;
						SET @phraseAs = CONCAT(N' AS ', @phraseAS);
					END;
						
					SET @errorNumber = 0;
					BEGIN TRY
						SET @query = CONCAT(@grVerb, ' ', @grPermission, ' ', @grOnObject, @onObj, ' TO [', @linFullLoginName, ']', @grPerModifier2, @phraseAS, ';');
						IF @linActiveNow = N'Yes' EXEC sp_executesql @query;
					END TRY
					BEGIN CATCH
						SELECT @errorNumber = ERROR_NUMBER(), @errorMessage = ERROR_MESSAGE(), @errorLine = ERROR_LINE();
					END CATCH
					PRINT @query;
					IF NOT @errorNumber = 0 PRINT CONCAT(N'ERROR ', @errorNumber, ', rolLineNo ', @linLineNo, ', GrLineNo ', @grLineNo, ', ', @errorMessage);

					FETCH NEXT FROM grant_cursor 
						INTO @grLineNo, @grRole, @grScope, @grVerb, @linDatabase, @grPermission, @grOnObject, @grObjectName, @grGrantOpt, @grAS;
				END;

				CLOSE grant_cursor;
				DEALLOCATE grant_cursor;
			END;
		END;

		IF ( @linOpIVerb = N'ALTER')
		BEGIN
			IF EXISTS (SELECT [sys].[sql_logins].[name], [sys].[sql_logins].[type] 
				FROM [sys].[sql_logins] 
				WHERE [type] IN (N'U', N'S') AND [name] = @linLoginName) 
			BEGIN
				/* Assemble Query component @status_options */
				SET @status_options = NULL;
				IF @linLoginStatus IS NOT NULL SET @status_options = @linLoginStatus;

				/* Assemble Query component @linLoginName, adding domain for Windows */
				IF (@linOpILoginWindows = N'Windows' AND @linLoginDomain IS NOT NULL) SET @linLoginName = CONCAT(@linLoginDomain, N'\', @linLoginName);

				/* Assemble Query component @cryptographic_credentials_option */
				SET @cryptographic_credentials_option = NULL;
				IF @linEKMCredentialVerb = N'ADD EKMCREDENTIAL' SET @linEKMCredentialVerb = N'ADD CREDENTIAL ';
				IF @linEKMCredentialVerb = N'DROP EKMCREDENTIAL' SET @linEKMCredentialVerb = N'DROP CREDENTIAL ';
				IF @linEKMCredentialVerb IS NOT NULL SET @cryptographic_credentials_option = CONCAT(@linEKMCredentialVerb, @linEKMCredentialName);

				/* Assemble Query component @defaultLanguageDB */
				SET @defaultLanguageDB = NULL;
				IF @linDefaultDatabase IS NOT NULL SET @defaultLanguageDB = CONCAT(@defaultLanguageDB, N'DEFAULT_DATABASE = ', @linDefaultDatabase);
				IF @defaultLanguageDB IS NOT NULL AND @linDefaultLanguage IS NOT NULL SET @defaultLanguageDB = @defaultLanguageDB + N', ';
				IF @linDefaultLanguage IS NOT NULL SET @defaultLanguageDB = CONCAT(@defaultLanguageDB, N'DEFAULT_LANGUAGE = ', @linDefaultLanguage);

				/* Assemble Query component @set_options for SQL Server */
				SET @set_options = NULL;
				IF @linOpILoginWindows IN(N'SQLServer', N'SvrCredential')
				BEGIN
					IF @linSetPassword IS NOT NULL SET @set_options = CONCAT(N'PASSWORD = ''', @linSetPassword, N'''');
					IF @set_options IS NOT NULL AND @linHashed =N'Yes' set @set_options = CONCAT(@set_options, N' HASHED');

					IF @set_options IS NOT NULL AND @linOldPassword IS NOT NULL SET @set_options = @set_options + N' ';
					IF @linOldPassword IS NOT NULL SET @set_options = CONCAT(@set_options, N'OLD PASSWORD = ''', @linOldPassword, N'''');
					IF @set_options IS NOT NULL AND @linOldPassword IS NOT NULL AND @linMustChange IN(N'MUST_CHANGE', N'UNLOCK') SET @set_options = @set_options + N' ';
					IF @linOldPassword IS NOT NULL AND @linMustChange IN(N'MUST_CHANGE', N'UNLOCK') SET @set_options = CONCAT(@set_options, N' ', @linMustChange);

					IF @set_options IS NOT NULL AND @linNewLoginName IS NOT NULL SET @set_options = @set_options + N', ';
					IF @linNewLoginName IS NOT NULL SET @set_options = CONCAT(@set_options, N' NAME = ', @linNewLoginName);
					
					IF @set_options IS NOT NULL AND @linCheckPolicy IN(N'ON', N'OFF') SET @set_options = @set_options + ', ';
					IF @linCheckPolicy IN(N'ON', N'OFF') SET @set_options = CONCAT(@set_options, N' CHECK_POLICY = ', @linCheckPolicy);

					IF @set_options IS NOT NULL AND @linPasswordExpires IN(N'ON', N'OFF') SET @set_options = @set_options + ', ';
					IF @linPasswordExpires IN(N'ON', N'OFF') SET @set_options = CONCAT(@set_options, N' CHECK_EXPIRATION = ', @linPasswordExpires);

					IF @set_options IS NOT NULL AND @linSvrCredential IS NOT NULL SET @set_options = @set_options + ', ';
					IF @linSvrCredential IS NOT NULL SET @set_options = CONCAT(@set_options, N' CREDENTIAL = ', @linSvrCredential);
				
					IF @set_options IS NOT NULL AND @linSvrCredentialVerb = N'NO CREDENTIAL' SET @set_options = @set_options + ', ';
					IF @linSvrCredentialVerb = N'NO CREDENTIAL' SET @set_options = CONCAT(@set_options, @linSvrCredentialVerb);
				END;

				/* Assemble the ALTER query from its components */
				
				SET @query = NULL
				IF @status_options IS NOT NULL SET @query = @status_options;
				IF @query IS NOT NULL AND @set_options IS NOT NULL SET @query = @query + N', ';
				IF @set_options IS NOT NULL SET @query = CONCAT(@query, N' WITH ', @set_options);
				IF @query IS NOT NULL AND @defaultLanguageDB IS NOT NULL SET @query = @query + N', ';
				IF @defaultLanguageDB IS NOT NULL set @query = CONCAT(@query, N' ', @defaultLanguageDB);
				IF @query IS NOT NULL AND @cryptographic_credentials_option IS NOT NULL SET @query = @query + N' ';
				IF @cryptographic_credentials_option IS NOT NULL SET @query = CONCAT(@query, N' ', @cryptographic_credentials_option);
				SET @query = CONCAT(N'ALTER LOGIN [', @linLoginName, N'] ', @query, N';');
				
				/* Execute the ALTER query */
				SET @errorNumber = 0;
				PRINT N'';
				PRINT @query;
				BEGIN TRY
					IF @linActiveNow = N'Yes' EXEC sp_executesql @query;
				END TRY
				BEGIN CATCH
					SELECT @errorNumber = ERROR_NUMBER(), @errorMessage = ERROR_MESSAGE(), @errorLine = ERROR_LINE();
				END CATCH
				/* Query was already printed to log creation. */
				IF NOT @errorNumber = 0 PRINT CONCAT(N'ERROR ', @errorNumber, ', Line ', @errorLine, ' ', @errorMessage);
			END;
			ELSE PRINT CONCAT(@linOpILoginWindows, N' Login Name ', @linLoginName, N'does not exist, cannot ALTER');
		END;

		IF ( @linOpIVerb = N'ADD/UPDATE' OR @linOpIVerb = N'ALTER')
		BEGIN
			/* Add Server Login GRANT * and * FIXED OpI Server-Role membership */
			/* First check that the Server-Login Name has been created */
			IF EXISTS (SELECT [sys].[server_principals].[name], [sys].[server_principals].[type] 
				FROM [sys].[server_principals] 
				WHERE [type] IN (N'U', N'S') AND [name] = @linLoginName) 
			BEGIN
				IF (@linRoleVerb = N'ADD' OR @linRoleVerb = N'DROP')
				BEGIN
					/* ADD/DROP membership to GRANT-Role Permissions Role for the server if that role exists */
					IF EXISTS (SELECT [sys].[server_principals].[name], [sys].[server_principals].[type] 
						FROM [sys].[server_principals] 
						WHERE [type] IN (N'R') AND [name] = @linRoleName)
					BEGIN
						SET @errorNumber = 0;
						BEGIN TRY
							SET @query = CONCAT(N'ALTER SERVER ROLE ', @linRoleName, ' ', @linRoleVerb, N' MEMBER ', @linFullLoginName, N';');
							IF @linActiveNow = N'Yes' EXEC sp_executesql @query;
						END TRY
						BEGIN CATCH
							SELECT @errorNumber = ERROR_NUMBER(), @errorMessage = ERROR_MESSAGE(), @errorLine = ERROR_LINE();
						END CATCH
						PRINT @query;
						IF NOT @errorNumber = 0 PRINT CONCAT(N'ERROR ', @errorNumber, ', Line ', @errorLine, ' ', @errorMessage);
					END;
				END;
			END;
		END;
	END;

	FETCH NEXT FROM logincreate_query_cursor 
		INTO @linLineNo, @linActiveNow, @linOpIVerb, @linOpILoginWindows, @linLoginDomain, @linLoginName, @linIdentifierText, @linIdentifierToken, @linSetPassword, @linHashed, @linMustChange, @linPasswordExpires, @linCheckPolicy, @linRoleVerb, @linRoleName, @linSID, @linSvrCredentialVerb, @linSvrCredential, @linWinCertificate, @linAsymmetricKey, @linDefaultDatabase, @linDefaultLanguage, @linLoginStatus, @linOldPassword, @linNewLoginName, @linEKMCredentialVerb, @linEKMCredentialName, @linOnEnvironment, @linDatabase, @linScope, @linFullLoginName;
END;

CLOSE logincreate_query_cursor;
DEALLOCATE logincreate_query_cursor;

/* LICENSING
Power Operational Intelligence ("PowerOpI") Framework
   Copyright (C) 2020 Richard M. Bixler
Derived products MSAccessTool, PowerQueryTool, SQLServerTool, and SQLServerRemote
   Copyright  2021 Richard M. Bixler
Derived product OpISQLServerSecurity
   Copyright  2022 Richard M. Bixler

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or 
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

Contact:
rmbpublic@comcast.net
www.softtoyssoftware.com */